from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, cast

from .utils import current_manager, to_coroutinefn

if TYPE_CHECKING:
    from _compat import Concatenate, Unpack

    from .manager import Donald
    from .types import TInterval, TRunArgs, TTaskParams, TTaskParamsPartial


class TaskWrapper:
    """Wrap a given function into a Task object."""

    __slots__ = ("_failback", "_fn", "_manager", "_params")

    def __init__(self, manager: Donald | None, fn: Callable, /,  # noqa: PLR0913
                 bind: bool = False, delay: float = 0, timeout: float = 0,  # noqa: FBT001, FBT002
                 retries_max: int = 0, retries_backoff_factor: float = 0,
                 retries_backoff_max: float = 600):
        assert "<locals>" not in fn.__qualname__, "Can't use local functions as tasks"

        self._manager = manager
        self._fn = to_coroutinefn(fn)
        self._failback: Callable | None = None
        self._params: TTaskParams = {
            "bind": bind,
            "delay": delay,
            "timeout": timeout,
            "retries_max": retries_max,
            "retries_backoff_max": retries_backoff_max,
            "retries_backoff_factor": retries_backoff_factor,
        }

    def __repr__(self):
        return f"<TaskWrapper {self._fn.__qualname__}>"

    def __call__(self, *args, **kwargs):
        return self._fn(*args, **kwargs)

    @staticmethod
    def import_path(fn: Callable) -> str:
        return f"{fn.__module__}.{fn.__qualname__}"

    def get_run(self, *args, kwargs: dict, **params: Unpack[TTaskParamsPartial]) -> TaskRun:
        task_params = cast("TTaskParams", dict(self._params, **params))
        return TaskRun(self.import_path(self._fn), args, kwargs or {}, task_params)

    def submit(self, *args, **kwargs):
        return self.apply_submit(*args, kwargs=kwargs)

    async def submit_and_wait(self, *args, timeout=10, **kwargs):
        manager = self._manager
        if not manager:
            raise RuntimeError("Manager is not set for this task")
        run = self.get_run(*args, kwargs=kwargs or {})
        return await manager.submit_and_wait(run, timeout=timeout)

    def apply_submit(self, *args, kwargs: dict | None = None, **params):
        manager = self._manager
        if not manager:
            raise RuntimeError("Manager is not set for this task")
        run = self.get_run(*args, kwargs=kwargs or {}, **params)
        return manager.submit(run)

    def schedule(self, interval: TInterval):
        manager = self._manager
        if not manager:
            raise RuntimeError("Manager is not set for this task")
        return manager.schedule(interval)(self)

    def failback(self):
        def wrapper(fn: Callable[Concatenate[Exception, ...], Any]):
            self._failback = to_coroutinefn(fn)
        return wrapper


class TaskRun:
    """Wrap a given function into a Task object."""

    __slots__ = ("data", "retries")

    def __init__(
        self,
        path: str,
        args: tuple,
        kwargs: dict,
        params: TTaskParams,
    ):
        if params.get("bind"):
            args = (self, *args)

        self.retries = 0
        self.data: TRunArgs = (path, args, kwargs, params)

    def retry(self: TaskRun):
        manager: Donald = current_manager.value
        self.retries += 1
        return manager.submit(self)

    def __repr__(self):
        path = self.data[0]
        return f"<TaskRun {path}>"



from __future__ import annotations

import asyncio
from asyncio import sleep
from asyncio.locks import Event
from asyncio.tasks import create_task, gather
from datetime import timedelta
from typing import TYPE_CHECKING, Callable, Coroutine, Optional

from crontab import CronTab

from donald.tasks import TaskWrapper

from .utils import SchedulerNotReadyError, logger

if TYPE_CHECKING:
    from donald.types import TInterval


class Scheduler:
    """Manage periodic and cron-based task schedules."""

    def __init__(self) -> None:
        self._schedule: list[Callable[[], Coroutine]] = []
        self._tasks: list[asyncio.Task] = []
        self._finished: Optional[Event] = None

    def start(self):
        """Start all scheduled tasks."""
        logger.info("Starting task scheduler")
        self._finished = Event()
        for scheduler in self._schedule:
            self._tasks.append(create_task(scheduler()))

    async def stop(self):
        """Stop all scheduled tasks."""
        for task in self._tasks:
            task.cancel()
        await self.join()
        if self._finished is not None:
            self._finished.set()

    def wait(self):
        """Wait for scheduler to finish."""
        if self._finished is None:
            raise SchedulerNotReadyError
        return self._finished.wait()

    async def join(self, timeout: float | None = None):
        """Wait for all scheduler tasks to complete with optional timeout."""
        try:
            await asyncio.wait_for(
                gather(*self._tasks, return_exceptions=True),
                timeout=timeout,
            )
        except asyncio.TimeoutError:
            logger.warning("Scheduler join timed out")
        self._tasks.clear()

    def schedule(  # noqa: C901
            self, interval: TInterval, *, run_immediately: bool = False, backoff: float = 0
        ) -> Callable[[TaskWrapper], TaskWrapper]:
        """
        Schedule a task to run periodically with optional immediate run and backoff on failure.
        """

        assert not callable(interval), "Use @manager.schedule(interval)"

        if isinstance(interval, str):
            interval = CronTab(interval)

        if isinstance(interval, timedelta):
            timer = interval.total_seconds

        elif isinstance(interval, CronTab):
            timer = lambda: interval.next(default_utc=True)  # type: ignore[union-attr]

        else:
            timer = lambda: float(interval)  # type: ignore[arg-type]

        def wrapper(task: TaskWrapper):
            assert isinstance(task, TaskWrapper), "Only tasks can be scheduled."

            async def submit():
                """Submit the task to the event loop."""
                try:
                    task.submit()
                except Exception as exc:  # noqa: BLE001
                    logger.exception(
                        "Immediate run of '%s' failed",
                        task.import_path(task._fn), exc_info=exc)
                    if backoff:
                        await sleep(backoff)

            async def scheduler():
                if run_immediately:
                    await submit()

                while True:
                    to_sleep = max(timer(), 1e-2)
                    logger.info("Next '%s' in %0.2f s", task.import_path(task._fn), to_sleep)
                    await sleep(to_sleep)
                    await submit()

            self._schedule.append(scheduler)
            return task

        return wrapper



import pytest

from donald import Donald
from donald.manager import TaskWrapper


def test_base():
    donald = Donald()
    assert donald
    assert donald._backend


def foo():
    return 42


async def test_wrap_task():
    donald = Donald()

    tw = donald.task()(foo)

    assert tw
    assert tw.import_path(tw._fn) == "tests.test__base.foo"
    assert isinstance(tw, TaskWrapper)
    assert tw._fn
    assert await tw() == 42

    tw = donald.task(delay=10)(foo)

    assert tw
    assert isinstance(tw, TaskWrapper)
    assert await tw() == 42
    assert tw._params["delay"] == 10


async def test_dont_wrap_local_tasks():
    donald = Donald()

    def foo():
        return 1

    with pytest.raises(AssertionError):
        donald.task()(foo)



import pytest

from .tasks import async_task, bind_task, manager, nested_task


@pytest.fixture(autouse=True)
async def donald():
    async with manager:
        w1 = manager.create_worker()
        w2 = manager.create_worker()
        w1.start()
        w2.start()
        yield manager
        await w1.stop()
        await w2.stop()


async def test_submit_task(sleep, checklog):
    res = async_task.submit()
    assert await res

    await sleep(1e-1)

    assert checklog("Run async_task 42")

    for n in range(3):
        assert await async_task.submit(n)

    await sleep(1e-1)
    assert checklog("Run async_task 0")
    assert checklog("Run async_task 1")
    assert checklog("Run async_task 2")


async def test_delay(sleep, checklog):
    res = async_task.apply_submit(delay=2e-1)
    assert await res

    await sleep(1e-1)
    assert not checklog("Run async_task 42")
    await sleep(2e-1)

    assert checklog("Run async_task 42")


async def test_timeout(sleep, checklog):
    await async_task.apply_submit(timeout=1e-1, kwargs={"timeout": 2e-1})
    await sleep(3e-1)
    assert not checklog("Run async_task 42")


async def test_nested_tasks(sleep, checklog):
    nested_task.submit()
    await sleep(3e-2)
    assert checklog("Run async_task nested", min_count=3)


async def test_binded_task(sleep, checklog):
    bind_task.submit()
    await sleep()
    assert checklog("Run bind_task TaskRun")


async def test_submit_and_wait(donald, sleep, checklog):
    res = await async_task.submit_and_wait()
    assert res == 42

    assert checklog("Run async_task 42")


async def test_health_check(donald, sleep, checklog):
    res = await donald.healthcheck()
    assert res is True



from __future__ import annotations

import os
from asyncio import iscoroutine
from asyncio.exceptions import CancelledError
from asyncio.locks import Event, Semaphore
from asyncio.tasks import Task, create_task, gather, sleep
from contextlib import suppress
from importlib import metadata
from random import random
from typing import TYPE_CHECKING, AsyncIterator, ClassVar, cast

from donald.tasks import TaskRun, TaskWrapper

from ._compat import async_timeout
from .utils import import_obj, logger

if TYPE_CHECKING:

    from .backend import BaseBackend
    from .types import TRunArgs, TTaskParams, TWorkerParams

BANNER = r"""

$$$$$$$\                                $$\       $$\
$$  __$$\                               $$ |      $$ |
$$ |  $$ | $$$$$$\  $$$$$$$\   $$$$$$\  $$ | $$$$$$$ |
$$ |  $$ |$$  __$$\ $$  __$$\  \____$$\ $$ |$$  __$$ |
$$ |  $$ |$$ /  $$ |$$ |  $$ | $$$$$$$ |$$ |$$ /  $$ |
$$ |  $$ |$$ |  $$ |$$ |  $$ |$$  __$$ |$$ |$$ |  $$ |
$$$$$$$  |\$$$$$$  |$$ |  $$ |\$$$$$$$ |$$ |\$$$$$$$ |
\_______/  \______/ \__|  \__| \_______|\__| \_______|
"""


class Worker:
    defaults: ClassVar[TWorkerParams] = {
        "max_tasks": 0,
        "task_defaults": None,
        "on_start": None,
        "on_stop": None,
        "on_error": None,
        "show_banner": False,
    }

    def __init__(self, backend: BaseBackend, params: TWorkerParams):
        self._backend = backend
        self._runner = None
        self._params = cast("TWorkerParams", dict(self.defaults, **params))
        self._task_params = cast("TTaskParams", self._params["task_defaults"] or {})

        self.on_error = self._params.get("on_error")

        max_tasks = self._params["max_tasks"]
        self._sem = (max_tasks and Semaphore(max_tasks)) or None

        self._tasks: set[Task] = set()
        self._finished = Event()

    def start(self):
        """Start the worker."""
        msg = (self._params.get("show_banner") and BANNER) or ""
        msg += f"\n\nDonald v{metadata.version('donald')} - Worker"
        msg += f"\nBackend: {self._backend.backend_type}"
        msg += f"\nPID: {os.getpid()}\n"

        logger.info(msg)
        self._finished.clear()
        self._runner = create_task(self.run_worker(), name="Worker Run Task")
        self._runner.add_done_callback(self.finish_runner)

    async def stop(self):
        """Stop the worker."""
        logger.info("Stopping worker")
        for task in self._tasks:
            task.cancel()

        if self._runner and not self._runner.done():
            self._runner.add_done_callback(lambda _: self._finished.set())
            self._runner.cancel()

        await self.join()

        on_stop = self._params.get("on_stop")
        if on_stop:
            await on_stop()

        self._finished.set()

    def wait(self):
        return self._finished.wait()

    async def run_worker(self: Worker):
        on_start = self._params.get("on_start")
        if on_start:
            await on_start()

        logger.info("Worker started")
        task_iter: AsyncIterator[TRunArgs] = await self._backend.subscribe()
        sem, tasks, finish_task = self._sem, self._tasks, self.finish_task
        async for task_msg in task_iter:
            path, args, kwargs, params = task_msg
            try:
                tw = import_obj(path)
            except Exception as exc:  # noqa: BLE001
                logger.exception("Failed to get task: %s", path, exc_info=exc)
                continue

            if not isinstance(tw, TaskWrapper):
                tw = TaskWrapper(None, tw)

            task_params = cast("TTaskParams", dict(self._task_params, **params))
            task = create_task(self.run_task(tw, args, kwargs, **task_params), name=path)
            tasks.add(task)
            task.add_done_callback(finish_task)
            logger.info("Run: '%s' (%d)", path, id(task))
            if sem:
                await sem.acquire()

    async def run_task(  # noqa: PLR0913
        self,
        tw: TaskWrapper,
        args: tuple,
        kwargs: dict,
        *,
        retries: int = 0,
        bind: bool = False,
        delay: float = 0,
        timeout: float = 0,
        retries_max: int = 0,
        retries_backoff_max: float = 600,
        retries_backoff_factor: float = .5,
        reply_to: str | None = None,
        correlation_id: str | None = None,
    ):
        """Run a task with the given parameters."""
        # Process delay
        if delay:
            await sleep(cast("float", delay))

        # Process timeout
        try:
            if timeout:
                async with async_timeout(cast("float", timeout)):
                    result = await tw._fn(*args, **kwargs)
            else:
                result = await tw._fn(*args, **kwargs)

            if reply_to and correlation_id:
                await self._backend.callback(result, reply_to, correlation_id)

            return result

        except Exception as exc:
            if bind:
                task_run: TaskRun = args[0]
                task_run.retries += 1
                retries = task_run.retries
            else:
                retries += 1

            if retries <= retries_max:
                backoff = min(retries_backoff_max, retries_backoff_factor * (2 ** (retries - 1) + random()))  # noqa: S311, E501
                logger.info("Retry: '%s' (%d) in %.2fs", tw._fn.__qualname__, retries, backoff)
                return await self.run_task(
                    tw,
                    args,
                    kwargs,
                    retries=retries,
                    bind=bind,
                    delay=delay + backoff,
                    timeout=timeout,
                    retries_max=retries_max,
                    retries_backoff_factor=retries_backoff_factor,
                    reply_to=reply_to,
                    correlation_id=correlation_id,
                )

            if tw._failback:
                return await tw._failback(exc, *args, **kwargs)

            raise

    def finish_task(self, task: Task):
        """Callback to handle the completion of a task."""
        self._tasks.discard(task)
        if task.cancelled():
            logger.info("Task cancelled: '%s'", task.get_name())

        with suppress(CancelledError):
            exc = task.exception()
            if exc:
                logger.exception("Fail: '%s' (%d)", task.get_name(), id(task), exc_info=exc)
                if self.on_error:
                    coro = self.on_error(exc)
                    if iscoroutine(coro):
                        create_task(coro)

        if self._sem:
            self._sem.release()

        logger.info("Finished: '%s' (%d)", task.get_name(), id(task))

    def finish_runner(self, runner: Task):
        """Callback to handle the completion of the worker runner."""
        with suppress(CancelledError):
            exc = runner.exception()
            if exc:
                logger.exception("Worker runner failed", exc_info=exc)
            else:
                logger.info("Worker runner finished successfully")
        self._finished.set()

    async def join(self):
        """Wait for all tasks to complete."""
        tasks = self._tasks
        if tasks:
            logger.info("Waiting for %d tasks to complete", len(self._tasks))
            return await gather(*tasks, return_exceptions=True)
        return None



from __future__ import annotations

import logging
from asyncio.coroutines import iscoroutinefunction
from functools import wraps
from importlib import import_module
from threading import local
from typing import TYPE_CHECKING, Any, Callable, Coroutine, Final, overload

if TYPE_CHECKING:
    from donald.types import TV, TVAsyncFn


def import_obj(path: str) -> Any:
    """Import an object from a string path."""
    module_path, _, obj_path = path.rpartition(".")
    try:
        module = import_module(module_path)
    except (ImportError, AttributeError, ValueError) as exc:
        raise ImportError(f"Could not import {path!r}") from exc
    else:
        return getattr(module, obj_path)


@overload
def to_coroutinefn(fn: TVAsyncFn) -> TVAsyncFn:
    ...


@overload
def to_coroutinefn(fn: Callable[..., TV]) -> Callable[..., Coroutine[Any, Any, TV]]:
    ...


def to_coroutinefn(fn):
    """Convert a function to a coroutine function."""
    if iscoroutinefunction(fn):
        return fn

    @wraps(fn)
    async def corofn(*args, **kwargs):
        return fn(*args, **kwargs)

    return corofn


class DonaldError(RuntimeError):
    """Donald error."""


class ManagerNotReadyError(DonaldError):
    """Manager is not ready."""


class BackendNotReadyError(DonaldError):
    """Backend is not ready."""


class SchedulerNotReadyError(DonaldError):
    """Scheduler is not ready."""


logger: Final = logging.getLogger("donald")

current_manager: Final = local()
current_manager.value = None



from __future__ import annotations

from asyncio import Queue, sleep
from pickle import dumps, loads
from typing import TYPE_CHECKING, Any, AsyncIterator, ClassVar, Mapping
from uuid import uuid4

from aio_pika import DeliveryMode, Exchange, Message, connect_robust

from ._compat import async_timeout
from .utils import BackendNotReadyError, logger

if TYPE_CHECKING:
    from aio_pika import Channel
    from redis.asyncio import Redis

    from donald.types import TTaskParams

    from .types import TBackendType, TRunArgs


class BaseBackend:
    backend_type: TBackendType
    defaults: ClassVar[Mapping] = {}

    def __init__(self, params: Mapping):
        self.is_connected = False
        self.params = dict(self.defaults, **params)
        self.__backend__: Any = None

    async def connect(self):
        logger.info("Connecting to Tasks Backend: %s", self.backend_type)
        await self._connect()
        self.is_connected = True
        logger.info("Connected to Tasks Backend: %s", self.backend_type)

    async def disconnect(self):
        logger.info("Disconnecting from Tasks Backend: %s", self.backend_type)
        self.is_connected = False
        await self._disconnect()
        logger.info("Disconnected from Tasks Backend: %s", self.backend_type)

    def submit(self, data: TRunArgs):
        if not self.is_connected:
            raise BackendNotReadyError
        _data = dumps(data)
        return self._submit(_data)

    async def submit_and_wait(self, data: TRunArgs, timeout=10):
        if not self.is_connected:
            raise BackendNotReadyError

        return await self._submit_and_wait(data, timeout=timeout)

    async def callback(self, result, reply_to: str, correlation_id: str):
        raise NotImplementedError

    async def subscribe(self) -> AsyncIterator[TRunArgs]:
        raise NotImplementedError

    async def _connect(self):
        raise NotImplementedError

    async def _disconnect(self):
        pass

    async def _submit(self, data):
        raise NotImplementedError

    async def _submit_and_wait(self, data, timeout=10):
        raise NotImplementedError


class MemoryBackend(BaseBackend):
    backend_type: TBackendType = "memory"

    @property
    def rx(self) -> Queue:
        if self.__backend__ is None:
            raise BackendNotReadyError

        return self.__backend__

    async def _connect(self):
        self.__backend__ = Queue()
        self.__results__ = Queue()

    async def _submit(self, data):
        self.rx.put_nowait(data)
        return True

    async def subscribe(self):
        await self.connect()

        async def iter_tasks():
            while self.is_connected:
                msg = await self.rx.get()
                if msg is None:
                    continue

                yield loads(msg)
                await sleep(0)

        return iter_tasks()

    async def callback(self, result, reply_to: str, correlation_id: str):
        return self.__results__.put_nowait(
            dumps({"result": result, "correlation_id": correlation_id})
        )

    async def _submit_and_wait(self, data: TRunArgs, timeout=10):
        correlation_id = str(uuid4())
        task_name, args, kwargs, params = data
        _params: TTaskParams = {**params, "correlation_id": correlation_id, "reply_to": "results"}
        data = (task_name, args, kwargs, _params)
        await self._submit(dumps(data))

        try:
            async with async_timeout(timeout):
                while True:
                    msg = await self.__results__.get()
                    res = loads(msg)
                    if res.get("correlation_id") == correlation_id:
                        return res.get("result")
                    else:
                        # Put back unmatched results for other waits
                        self.__results__.put_nowait(msg)
        except TimeoutError as exc:
            raise TimeoutError("Task result timeout") from exc


class RedisBackend(BaseBackend):
    backend_type: TBackendType = "redis"
    defaults: ClassVar = {
        "url": "redis://localhost:6379/0",
        "channel": "donald",
        "consumer": "consumer-1",
    }

    def __init__(self, params: Mapping):
        self.is_connected = False
        self.params = dict(self.defaults, **params)
        self.__backend__: Any = None

    @property
    def redis(self) -> "Redis":
        if self.__backend__ is None:
            raise BackendNotReadyError
        return self.__backend__

    async def _connect(self):
        from redis.asyncio import from_url

        self.__backend__ = from_url(self.params["url"])

        stream = f"{self.params['channel']}:stream"
        group = f"{self.params['channel']}:group"

        try:
            await self.redis.xgroup_create(stream, group, id="$", mkstream=True)
            logger.info("Created Redis stream group '%s'", group)
        except Exception:  # noqa: BLE001
            logger.info("Redis stream group '%s' already exists", group)

        self.is_connected = True

    async def _disconnect(self):
        await self.redis.close()
        self.is_connected = False

    async def _submit(self, data):
        stream = f"{self.params['channel']}:stream"
        return await self.redis.xadd(stream, {"data": data})

    async def subscribe(self) -> AsyncIterator[TRunArgs]:
        stream = f"{self.params['channel']}:stream"
        group = f"{self.params['channel']}:group"
        consumer = self.params["consumer"]

        async def iter_tasks():
            while self.is_connected:
                msgs = await self.redis.xreadgroup(
                    group, consumer, streams={stream: ">"}, count=1, block=1000
                )
                if not msgs:
                    continue

                for _, entries in msgs:
                    for msg_id, msg_data in entries:
                        data = msg_data.get(b"data")
                        if data:
                            yield loads(data)
                            await self.redis.xack(stream, group, msg_id)

        return iter_tasks()

    async def callback(self, result, reply_to: str, correlation_id: str):
        payload = dumps(
            {
                "result": result,
                "correlation_id": correlation_id,
            }
        )
        # For Streams-based RPC, we can implement a separate stream for results
        stream = f"{reply_to}:stream"
        await self.redis.xadd(stream, {"data": payload})

    async def _submit_and_wait(self, data: TRunArgs, timeout=10):
        correlation_id = str(uuid4())
        task_name, args, kwargs, params = data
        params["correlation_id"] = correlation_id
        reply_to = f"{self.params['channel']}:rpc:{correlation_id}"
        params["reply_to"] = reply_to
        data = (task_name, args, kwargs, params)

        # Submit task
        await self._submit(dumps(data))

        # Wait for result on reply stream
        stream = f"{reply_to}:stream"
        try:
            async with async_timeout(timeout):
                while True:
                    msgs = await self.redis.xread({stream: "0"}, block=1000, count=1)
                    if not msgs:
                        continue
                    for _, entries in msgs:
                        for msg_id, msg_data in entries:
                            res = loads(msg_data[b"data"])
                            if res.get("correlation_id") == correlation_id:
                                await self.redis.xdel(stream, msg_id)
                                return res.get("result")
        except TimeoutError as exc:
            raise TimeoutError("Task result timeout") from exc


class AMQPBackend(BaseBackend):
    backend_type: TBackendType = "amqp"
    defaults: ClassVar = {
        "url": "amqp://guest:guest@localhost:5672/",
        "queue": "donald",
        "exchange": "",
    }

    def __init__(self, params: Mapping):
        super().__init__(params)
        self.__channel__: Channel | None = None
        self.__exchange__: Exchange | None = None

    async def _connect(self):
        self.__backend__ = await connect_robust(self.params["url"])
        self.__channel__ = channel = await self.__backend__.channel()

        await channel.set_qos(prefetch_count=1)
        self.__queue__ = await channel.declare_queue(self.params["queue"], durable=True)

        exchange_name = self.params["exchange"]
        if exchange_name:
            self.exchange = await channel.get_exchange(exchange_name)
        else:
            self.exchange = channel.default_exchange

    async def _disconnect(self):
        await self.__backend__.close()

    async def _submit(self, data, **params):
        return await self.exchange.publish(
            Message(
                body=data,
                message_id=str(uuid4()),
                delivery_mode=DeliveryMode.PERSISTENT,
                **params,
            ),
            routing_key=self.params["queue"],
        )

    async def subscribe(self) -> AsyncIterator[TRunArgs]:
        queue = self.__queue__

        async def iter_tasks() -> AsyncIterator[TRunArgs]:
            async with queue.iterator() as queue_iter:
                async for message in queue_iter:
                    async with message.process():
                        data: TRunArgs = loads(message.body)
                        if message.reply_to and message.correlation_id:
                            task_name, args, kwargs, params = data
                            _params: TTaskParams = {
                                **params,
                                "reply_to": message.reply_to,
                                "correlation_id": message.correlation_id,
                            }
                            data = (task_name, args, kwargs, _params)

                        yield data

        return iter_tasks()

    async def callback(self, result, reply_to: str, correlation_id: str):
        await self.exchange.publish(
            Message(
                body=dumps(result),
                correlation_id=correlation_id,
                delivery_mode=DeliveryMode.PERSISTENT,
            ),
            routing_key=reply_to,
        )

    async def _submit_and_wait(self, data: TRunArgs, timeout=10):
        channel = self.__channel__
        assert channel is not None, "Channel is not set"
        callback_queue = await channel.declare_queue(exclusive=True)

        correlation_id = str(uuid4())

        await self._submit(
            dumps(data),
            reply_to=callback_queue.name,
            correlation_id=correlation_id,
        )

        async with callback_queue.iterator() as queue_iter:
            try:
                async with async_timeout(timeout):
                    async for message in queue_iter:
                        async with message.process():
                            if message.correlation_id == correlation_id:
                                return loads(message.body)
            except TimeoutError as exc:
                raise TimeoutError("Task result timeout") from exc

    @property
    def exchange(self) -> Exchange:
        if self.__exchange__ is None:
            raise BackendNotReadyError
        return self.__exchange__

    @exchange.setter
    def exchange(self, value: Exchange):
        if not isinstance(value, Exchange):
            raise TypeError
        self.__exchange__ = value


BACKENDS: dict[TBackendType, type[BaseBackend]] = {
    "memory": MemoryBackend,
    "redis": RedisBackend,
    "amqp": AMQPBackend,
}

# ruff: noqa: S301, ARG002, PLC0415



from __future__ import annotations

import asyncio
import signal
from functools import wraps
from typing import TYPE_CHECKING, Awaitable, Callable

import click

from .utils import import_obj

if TYPE_CHECKING:
    from donald.manager import Donald

    from .types import TV


def import_manager(path: str) -> Donald:
    """Import a manager from a python path."""
    manager: Donald = import_obj(path)
    return manager


def process_await(fn: Callable[..., Awaitable[TV]]) -> Callable[..., TV]:
    @wraps(fn)
    @click.pass_context
    def wrapper(ctx, *args, **kwargs):
        loop = ctx.obj["loop"]
        return loop.run_until_complete(fn(ctx, *args, **kwargs))

    return wrapper


@click.group()
@click.option(
    "-M",
    "--manager",
    "manager",
    required=True,
    help="Python path to the manager",
)
@click.pass_context
def cli(ctx: click.Context, manager: str):
    ctx.obj["manager"] = import_manager(manager)


@cli.command(help="Launch a worker")
@click.option("-S", "--scheduler", "scheduler", is_flag=True, help="Start a scheduler")
@process_await
async def worker(ctx: click.Context, *, scheduler: bool = False, **params):
    """Launch a worker."""

    loop = ctx.obj["loop"]

    async def stop():
        loop.remove_signal_handler(signal.SIGTERM)
        loop.remove_signal_handler(signal.SIGINT)
        await worker.stop()
        if scheduler:
            await manager.scheduler.stop()
        await manager.stop()

    loop.add_signal_handler(signal.SIGINT, lambda: loop.create_task(stop()))
    loop.add_signal_handler(signal.SIGTERM, lambda: loop.create_task(stop()))

    manager: Donald = ctx.obj["manager"]
    await manager.start()
    if scheduler:
        manager.scheduler.start()

    worker = manager.create_worker(show_banner=True, **params)
    worker.start()

    await worker.wait()


@cli.command(help="Launch a scheduler")
@process_await
async def scheduler(ctx: click.Context):
    loop = ctx.obj["loop"]

    async def stop():
        loop.remove_signal_handler(signal.SIGTERM)
        loop.remove_signal_handler(signal.SIGINT)
        await manager.scheduler.stop()
        await manager.stop()

    loop.add_signal_handler(signal.SIGINT, lambda: loop.create_task(stop()))
    loop.add_signal_handler(signal.SIGTERM, lambda: loop.create_task(stop()))

    manager: Donald = ctx.obj["manager"]
    await manager.start()

    manager.scheduler.start()
    await manager.scheduler.wait()


def main():
    loop = asyncio.get_event_loop()
    cli(obj={"loop": loop})


if __name__ == "__main__":
    main()



import pytest

from .tasks import async_task, manager


@pytest.fixture
async def donald():
    manager.setup(backend="amqp")
    assert manager._backend
    assert manager._backend.params

    async with manager:
        w1 = manager.create_worker()
        w2 = manager.create_worker()
        w1.start()
        w2.start()
        yield manager
        await w1.stop()
        await w2.stop()


async def test_submit_task(donald, sleep, checklog):
    res = async_task.submit()
    assert res
    assert await res

    await sleep()
    assert checklog("Run async_task 42")

    for n in range(3):
        assert await async_task.submit(n)

    await sleep(1e-1)
    assert checklog("Run async_task 0")
    assert checklog("Run async_task 1")
    assert checklog("Run async_task 2")


async def test_submit_and_wait(donald, sleep, checklog):
    res = await async_task.submit_and_wait()
    assert res == 42

    assert checklog("Run async_task 42")


async def test_health_check(donald, sleep, checklog):
    res = await donald.healthcheck()
    assert res is True



try:
    from asyncio import timeout as async_timeout  # type: ignore[attr-defined]
except ImportError:  # python 310
    from async_timeout import timeout as async_timeout  # type: ignore[assignment, no-redef]


# ruff: noqa: F401



from __future__ import annotations

from asyncio.tasks import Task, create_task
from logging.config import dictConfig
from typing import TYPE_CHECKING, Any, Callable, ClassVar, cast

from typing_extensions import Unpack

from .backend import BACKENDS, BaseBackend
from .utils import ManagerNotReadyError, current_manager, logger
from .worker import Worker

if TYPE_CHECKING:
    from .types import (
        TInterval,
        TManagerParams,
        TTaskParamsPartial,
        TVWorkerOnErrFn,
        TVWorkerOnFn,
        TWorkerParams,
    )


class Donald:
    """Manage tasks and workers."""

    defaults: ClassVar[TManagerParams] = {
        "log_level": "INFO",
        "log_config": None,
        "backend": "memory",
        "backend_params": {},
        "worker_params": cast("TWorkerParams", {}),
    }

    _backend: BaseBackend

    def __init__(self: Donald, **params):
        self._params: TManagerParams = self.defaults
        self.is_started = False
        self.setup(**params)
        self.scheduler = Scheduler()
        self.submissions: set[Task] = set()
        current_manager.value = self

    def __repr__(self):
        return f"<Donald {self._params['backend']}>"

    @classmethod
    async def create(cls, **params) -> Donald:
        """Create a new instance of Donald and start it."""
        self = cls(**params)
        await self.start()
        return self

    def setup(self: Donald, **params):
        """Setup the manager."""
        assert not self.is_started, "Manager is already started"

        self._params = cast("TManagerParams", dict(self._params, **params))
        self._backend = BACKENDS[self._params["backend"]](
            self._params["backend_params"],
        )

        logger.setLevel(self._params["log_level"].upper())
        if self._params["log_config"]:
            dictConfig(self._params["log_config"])

    async def start(self):
        """Start the manager."""
        logger.info("Starting tasks manager")
        try:
            await self._backend.connect()
        except Exception as exc:
            logger.exception("Failed to connect backend", exc_info=exc)
            self.is_started = False
            raise
        else:
            self.is_started = True
            logger.info("Tasks manager started")

    async def stop(self):
        logger.info("Stopping tasks manager")
        try:
            await self._backend.disconnect()
        except Exception as e:  # noqa: BLE001
            logger.error(f"Error while disconnecting backend: {e}")
        finally:
            self.is_started = False
            logger.info("Tasks manager stopped")

    def create_worker(self: Donald, **params):
        """Create a worker."""
        worker_params = cast(
            "TWorkerParams",
            dict(self._params["worker_params"], **params),
        )
        return Worker(self._backend, worker_params)

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.stop()

    def schedule(self, interval: TInterval):
        """Schedule a task to the backend."""
        return self.scheduler.schedule(interval)

    def task(
        self,
        fn: Any = ...,
        /,
        **params: Unpack[TTaskParamsPartial],
    ) -> Callable[[Callable], TaskWrapper]:
        """Decorator to wrap a function into a Task object."""

        if fn is not ...:
            raise ValueError("Task decorator must be used with parentheses (e.g. @task() or @task(**params))") from None  # noqa: E501

        def wrapper(fn: Callable) -> TaskWrapper:
            return TaskWrapper(self, fn, **params)

        return wrapper

    def _assert_run(self, run: TaskRun | Callable, *args, **kwargs) -> TaskRun:
        """Assert that the run is a TaskRun or convert it to one."""
        if isinstance(run, TaskRun):
            return run

        elif callable(run):
            wrapper = TaskWrapper(self, run)
            return wrapper.get_run(*args, kwargs=kwargs)

        else:
            raise TypeError("run must be a TaskRun or a callable function")

    def submit(self, run: TaskRun | Callable, *args, **kwargs) -> Task:
        """Submit a task to the backend."""
        if not self.is_started:
            raise ManagerNotReadyError

        run = self._assert_run(run, *args, **kwargs)
        task = create_task(self._backend.submit(run.data))
        self.submissions.add(task)
        task.add_done_callback(self.submissions.discard)
        return task

    async def submit_and_wait(self, run: TaskRun | Callable, *args, timeout=10, **kwargs) -> Any:
        """Submit a task to the backend and wait for result (RPC)."""
        if not self.is_started:
            raise ManagerNotReadyError

        run = self._assert_run(run, *args, **kwargs)
        return await self._backend.submit_and_wait(run.data, timeout=timeout)

    def on_start(self, fn: TVWorkerOnFn) -> TVWorkerOnFn:
        """Register a function to be called on worker start."""
        self._params["worker_params"]["on_start"] = fn
        return fn

    def on_stop(self, fn: TVWorkerOnFn) -> TVWorkerOnFn:
        """Register a function to be called on worker stop."""
        self._params["worker_params"]["on_stop"] = fn
        return fn

    def on_error(self, fn: TVWorkerOnErrFn) -> TVWorkerOnErrFn:
        """Register a function to be called on worker error."""
        self._params["worker_params"]["on_error"] = fn
        return fn

    async def healthcheck(self, timeout=10) -> bool:
        try:
            result = await self.submit_and_wait(ping, timeout=timeout)
            return result == "pong"
        except Exception as exc:  # noqa: BLE001
            logger.exception("Healthcheck failed", exc_info=exc)
            return False


async def ping():
    """Ping the manager to check if it's running."""
    return "pong"

from .scheduler import Scheduler  # noqa: E402
from .tasks import TaskRun, TaskWrapper  # noqa: E402



from __future__ import annotations

from datetime import timedelta
from typing import (
    Any,
    Callable,
    Coroutine,
    Literal,
    Optional,
    TypedDict,
    TypeVar,
    Union,
)

from crontab import CronTab
from typing_extensions import NotRequired  # py310

TWorkerOnFn = Callable[[], Coroutine]
TVWorkerOnFn = TypeVar("TVWorkerOnFn", bound=TWorkerOnFn)
TWorkerOnErrFn = Callable[[BaseException], Coroutine]
TVWorkerOnErrFn = TypeVar("TVWorkerOnErrFn", bound=TWorkerOnErrFn)

TV = TypeVar("TV", bound=Any)
TAsyncFn = Callable[..., Coroutine]
TVAsyncFn = TypeVar("TVAsyncFn", bound=TAsyncFn)

TBackendType = Literal["memory", "redis", "amqp"]

TInterval = Union[timedelta, int, float, str, CronTab]

class TTaskParams(TypedDict):
    bind: bool
    delay: float
    timeout: float
    retries_max: int
    retries_backoff_factor: float
    retries_backoff_max: float
    reply_to: NotRequired[str]
    correlation_id: NotRequired[str]

class TTaskParamsPartial(TypedDict):
    bind: NotRequired[bool]
    delay: NotRequired[float]
    timeout: NotRequired[float]
    retries_max: NotRequired[int]
    retries_backoff_factor: NotRequired[float]
    retries_backoff_max: NotRequired[float]

TRunArgs = tuple[str, tuple, dict[str, Any], TTaskParams]

class TWorkerParams(TypedDict):
    max_tasks: Optional[int]
    on_start: Optional[TWorkerOnFn]
    on_stop: Optional[TWorkerOnFn]
    on_error: Optional[TWorkerOnErrFn]
    show_banner: Optional[bool]
    task_defaults: Optional[TTaskParamsPartial]


class TManagerParams(TypedDict):
    backend: TBackendType
    backend_params: dict
    log_config: dict | None
    log_level: str
    worker_params: TWorkerParams



from .tasks import manager


def test_base():
    assert manager.scheduler._schedule


async def test_schedule(checklog, sleep):
    async with manager:
        w = manager.create_worker()
        w.start()

        manager.scheduler.start()
        assert manager.scheduler._tasks

        await sleep(5e-1)

        await manager.scheduler.stop()
        assert not manager.scheduler._tasks

        await w.stop()

    assert checklog("Run scheduled_task 42", 4)
